from chellow.gas.engine import g_rates


DAILY_THRESHOLD = 145


def vb(ds):
    for hh in ds.hh_data:
        rate = float(
            g_rates(ds.sess, ds.caches, "ccl", hh["start_date"], True)[
                "ccl_gbp_per_kwh"
            ]
        )

        hh["ccl"] = rate



import csv
from decimal import Decimal
from io import StringIO

from utils import match

from chellow.e.computer import contract_func
from chellow.models import (
    BillType,
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    ReportRun,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_111 import _process_supply, content
from chellow.utils import ct_datetime, to_utc, utc_datetime


# End to end tests


def test_ete_error_message_for_invalid_bill_id(client, sess):
    query_string = {
        "bill_id": "0",
    }
    response = client.get("/reports/111", query_string=query_string)

    match(response, 404)


# HTTP level tests


def test_http_supplier_batch_with_mpan_cores(mocker, client, sess):
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    participant = Participant.insert(sess, "hhak", "AK Industries")
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    batch = supplier_contract.insert_batch(sess, "005", "batch 5")
    sess.commit()
    user = User.get_by_email_address(sess, "admin@example.com")
    MockThread = mocker.patch("chellow.reports.report_111.threading.Thread")

    query_string = {
        "batch_id": str(batch.id),
        "mpan_cores": "22 1065 3921 534",
    }
    response = client.get("/reports/111", query_string=query_string)

    match(response, 303)

    expected_args = (
        batch.id,
        None,
        None,
        None,
        None,
        user.id,
        ["22 1065 3921 534"],
        "_batch_005",
        1,
    )

    MockThread.assert_called_with(target=content, args=expected_args)


# Worker level tests


def test_process_supply(sess):
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill(ds):
    rate = 0.1
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['rate'] = {rate}
        bill_hh['off-rate'] = {0.1}
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * rate
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        valid_from,
        None,
        {},
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a b", "")
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "dd",
        "hjk",
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        bill_type,
        {"rate": [Decimal("0.1")]},
        supply,
    )
    report_run = ReportRun.insert(sess, "bill_check", None, "", {})
    sess.commit()

    report_context = {}
    supply_id = supply.id
    bill_ids = {bill.id}
    forecast_date = utc_datetime(2020, 7, 10)
    vbf = contract_func(report_context, supplier_contract, "virtual_bill")
    virtual_bill_titles = [
        "ccl-kwh",
        "ccl-rate",
        "ccl-gbp",
        "net-gbp",
        "vat-gbp",
        "gross-gbp",
        "sum-msp-kwh",
        "rate",
        "sum-msp-gbp",
        "problem",
    ]
    titles = [
        "batch",
        "bill-reference",
        "bill-type",
        "bill-kwh",
        "bill-net-gbp",
        "bill-vat-gbp",
        "bill-start-date",
        "bill-finish-date",
        "imp-mpan-core",
        "exp-mpan-core",
        "site-code",
        "site-name",
        "covered-from",
        "covered-to",
        "covered-bills",
        "metered-kwh",
    ]
    for t in virtual_bill_titles:
        titles.append("covered-" + t)
        titles.append("virtual-" + t)
        if t.endswith("-gbp"):
            titles.append("difference-" + t)

    f = StringIO()
    writer = csv.writer(f)

    _process_supply(
        sess,
        report_context,
        supply_id,
        bill_ids,
        forecast_date,
        supplier_contract,
        vbf,
        virtual_bill_titles,
        writer,
        titles,
        report_run,
    )
    expected = (
        "a b,hjk,N,10.00,10.00,10.00,2009-07-10 01:00,2009-07-10 01:00,"
        "22 7867 6232 781,,CI017,Water Works,2009-07-10 01:00,2009-07-10 01:00,"
        "1,0,,,,,,,0,10.0,0,10.0,10.0,,0,,,0,10.0,,0.1,0.1,,,0,,,"
        "virtual-off-rate,0.1\r\n"
    )
    assert f.getvalue() == expected


def test_content(sess, mocker):
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill(ds):
    rate = 0.1
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['rate'] = {rate}
        bill_hh['off-rate'] = {0.1}
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * rate
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        valid_from,
        None,
        {},
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a b", "")
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "dd",
        "hjk",
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        utc_datetime(2009, 7, 10),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        Decimal("10.00"),
        bill_type,
        {"rate": [Decimal("0.1")]},
        supply,
    )
    user_role = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "admin", user_role, None)
    report_run = ReportRun.insert(sess, "", user, "", {})
    sess.commit()

    virtual_bill_titles = [
        "ccl-kwh",
        "ccl-rate",
        "ccl-gbp",
        "net-gbp",
        "vat-gbp",
        "gross-gbp",
        "sum-msp-kwh",
        "rate",
        "sum-msp-gbp",
        "problem",
    ]
    titles = [
        "batch",
        "bill-reference",
        "bill-type",
        "bill-kwh",
        "bill-net-gbp",
        "bill-vat-gbp",
        "bill-start-date",
        "bill-finish-date",
        "imp-mpan-core",
        "exp-mpan-core",
        "site-code",
        "site-name",
        "covered-from",
        "covered-to",
        "covered-bills",
        "metered-kwh",
    ]
    for t in virtual_bill_titles:
        titles.append("covered-" + t)
        titles.append("virtual-" + t)
        if t.endswith("-gbp"):
            titles.append("difference-" + t)

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_111.open", return_value=mock_file)
    mocker.patch(
        "chellow.reports.report_111.chellow.dloads.make_names", return_value=("a", "b")
    )
    mocker.patch("chellow.reports.report_111.os.rename")

    content(
        batch.id,
        bill.id,
        supplier_contract.id,
        valid_from,
        valid_from,
        user.id,
        [],
        "",
        report_run.id,
    )



from datetime import datetime as Datetime
from decimal import Decimal

from openpyxl import Workbook

import pytest


from werkzeug.exceptions import BadRequest

from chellow.e.bill_parsers.engie_xlsx import _bd_add, _make_raw_bills, _parse_row
from chellow.utils import utc_datetime


def test_parse_row():
    wb = Workbook()
    ws = wb.active

    for column, value in enumerate(
        [
            "Power Comany Ltd.",
            "Bill Paja",
            "556",
            "BILL PAJA",
            "883",
            "1 True Way",
            "672770",
            Datetime(2019, 3, 31, 23, 30),
            Datetime(2019, 3, 31, 23, 30),
            "",
            "2019-03-01 - 2019-03-31",
            "",
            "",
            "",
            "Draft",
            "",
            "Product",
            "Hand Held Read -",
            Datetime(2019, 3, 31, 23, 30),
            Datetime(2019, 3, 31, 0, 0),
            "",
            "2299999999929",
            "",
            "",
            "",
            "785",
            "GBP",
            "INV",
            "",
            "",
        ],
        start=1,
    ):
        ws.cell(row=2, column=column, value=value)

    row = 2
    title_row = ["To Date"]

    bill = _parse_row(ws, row, title_row)
    assert bill["finish_date"] == utc_datetime(2019, 3, 31, 22, 30)


def test_bd_add():
    el_name = "duos_red"
    bd = {el_name: 0}
    val = None

    with pytest.raises(BadRequest):
        _bd_add(bd, el_name, val)


def test_bill_parser_engie_xls_billed_kwh():
    """
    Blank units should still give billed kWh
    """
    wb = Workbook()
    ws = wb.active

    row_vals = [
        "Mistral Wind Power Ltd.",
        "Bill Paja",
        "886572998",
        "Bill Paja",
        "869987122",
        "BA1 5TT",
        "99708221",
        Datetime(2019, 3, 31, 23, 30),
        42694,
        42629,
        "2016-08-01 - 2016-08-31",
        "458699",
        "Standard Deluxe",
        "Beautiful Breeze",
        "Accepted",
        "Pass Thru - UK Electricity Cost Component",
        "Product",
        "Renewables Obligation (RO)",
        Datetime(2019, 3, 31, 23, 30),
        Datetime(2019, 3, 31, 23, 30),
        "Commercial UK Energy VAT",
        "2298132107763",
        27997.33,
        "",
        0.0971123676,
        "9224",
        "GBP",
        "INV",
        "Renewables Obligation (RO)",
        "",
    ]
    for column, value in enumerate(row_vals, start=1):
        ws.cell(row=2, column=column, value=value)

    row = 2
    bill = _parse_row(ws, row, [])
    assert bill["kwh"] == Decimal("27997.33")


def test_make_raw_bills_vat():
    wb = Workbook()
    ws = wb.active

    row_vals = [
        "Mistral Wind Power Ltd.",
        "Bill Paja",
        "886572998",
        "Bill Paja",
        "869987122",
        "BA1 5TT",
        "99708221",
        Datetime(2019, 3, 31, 23, 30),
        Datetime(2019, 3, 31, 23, 30),
        Datetime(2019, 3, 31, 23, 30),
        "2016-08-01 - 2016-08-31",
        "",
        "",
        "",
        "Accepted",
        "",
        "Sales Tax",
        "Standard VAT@20%",
        "",
        "",
        "",
        "2298132107763",
        "",
        "",
        "",
        "9224",
        "GBP",
        "INV",
        "",
        "",
    ]

    for column, value in enumerate(row_vals, start=1):
        ws.cell(row=2, column=column, value=value)

    bills = _make_raw_bills(ws)
    expected_bills = [
        {
            "bill_type_code": "N",
            "kwh": Decimal("0"),
            "vat": Decimal("9224.00"),
            "net": Decimal("0.00"),
            "reads": [],
            "breakdown": {
                "vat_percentage": Decimal("20"),
            },
            "account": "22 9813 2107 763",
            "issue_date": utc_datetime(2019, 3, 31, 22, 30),
            "start_date": utc_datetime(2016, 7, 31, 23, 0),
            "finish_date": utc_datetime(2016, 8, 31, 22, 30),
            "mpan_core": "22 9813 2107 763",
            "reference": "99708221_2",
            "gross": Decimal("9224.00"),
        }
    ]
    assert bills == expected_bills



import os
import threading
import time
import traceback
from datetime import datetime as Datetime

from dateutil.relativedelta import relativedelta

from flask import g, request

import pytz

from sqlalchemy import Float, cast, func, or_
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null, true

import chellow.computer
import chellow.dloads
from chellow.models import Bill, Channel, Era, HhDatum, Session, Site, SiteEra, Supply
from chellow.utils import HH, hh_format, hh_max, hh_min, req_int
from chellow.views import chellow_redirect


def _content(sess, tmp_file, year, month, months, supply_id, user):
    supplies = (
        sess.query(Supply)
        .join(Era)
        .distinct()
        .options(joinedload(Supply.generator_type))
    )

    if supply_id is not None:
        supply = Supply.get_by_id(sess, supply_id)
        supplies = supplies.filter(Supply.id == supply.id)

    caches = {}

    start_date = Datetime(year, month, 1, tzinfo=pytz.utc) - relativedelta(
        months=months - 1
    )

    field_names = (
        "supply-name",
        "source-code",
        "generator-type",
        "month",
        "pc-code",
        "msn",
        "site-code",
        "site-name",
        "metering-type",
        "import-mpan-core",
        "metered-import-kwh",
        "metered-import-net-gbp",
        "metered-import-estimated-kwh",
        "billed-import-kwh",
        "billed-import-net-gbp",
        "export-mpan-core",
        "metered-export-kwh",
        "metered-export-estimated-kwh",
        "billed-export-kwh",
        "billed-export-net-gbp",
        "problem",
        "timestamp",
    )

    tmp_file.write("supply-id," + ",".join(field_names) + "\n")

    forecast_date = chellow.computer.forecast_date()

    for i in range(months):
        month_start = start_date + relativedelta(months=i)
        month_finish = month_start + relativedelta(months=1) - HH

        for supply in supplies.filter(
            Era.start_date <= month_finish,
            or_(Era.finish_date == null(), Era.finish_date >= month_start),
        ):
            generator_type = supply.generator_type
            if generator_type is None:
                generator_type = ""
            else:
                generator_type = generator_type.code

            source_code = supply.source.code
            eras = supply.find_eras(sess, month_start, month_finish)
            era = eras[-1]
            metering_type = era.meter_category

            site = (
                sess.query(Site)
                .join(SiteEra)
                .filter(SiteEra.era == era, SiteEra.is_physical == true())
                .one()
            )

            values = {
                "supply-name": supply.name,
                "source-code": source_code,
                "generator-type": generator_type,
                "month": hh_format(month_finish),
                "pc-code": era.pc.code,
                "msn": era.msn,
                "site-code": site.code,
                "site-name": site.name,
                "metering-type": metering_type,
                "problem": "",
            }

            tmp_file.write(str(supply.id) + ",")

            for is_import, pol_name in [(True, "import"), (False, "export")]:
                if is_import:
                    mpan_core = era.imp_mpan_core
                else:
                    mpan_core = era.exp_mpan_core

                values[pol_name + "-mpan-core"] = mpan_core
                kwh = 0
                est_kwh = 0

                if metering_type in ["hh", "amr"]:
                    est_kwh = (
                        sess.query(HhDatum.value)
                        .join(Channel)
                        .join(Era)
                        .filter(
                            HhDatum.status == "E",
                            Era.supply_id == supply.id,
                            Channel.channel_type == "ACTIVE",
                            Channel.imp_related == is_import,
                            HhDatum.start_date >= month_start,
                            HhDatum.start_date <= month_finish,
                        )
                        .first()
                    )
                    if est_kwh is None:
                        est_kwh = 0
                    else:
                        est_kwh = est_kwh[0]

                if not (is_import and source_code in ("net", "gen-net")):
                    kwh_sum = (
                        sess.query(cast(func.sum(HhDatum.value), Float))
                        .join(Channel)
                        .join(Era)
                        .filter(
                            Era.supply_id == supply.id,
                            Channel.channel_type == "ACTIVE",
                            Channel.imp_related == is_import,
                            HhDatum.start_date >= month_start,
                            HhDatum.start_date <= month_finish,
                        )
                        .one()[0]
                    )
                    if kwh_sum is not None:
                        kwh += kwh_sum

                values["metered-" + pol_name + "-estimated-kwh"] = est_kwh
                values["metered-" + pol_name + "-kwh"] = kwh
                values["metered-" + pol_name + "-net-gbp"] = 0
                values["billed-" + pol_name + "-kwh"] = 0
                values["billed-" + pol_name + "-net-gbp"] = 0
                values["billed-" + pol_name + "-apportioned-kwh"] = 0
                values["billed-" + pol_name + "-apportioned-net-gbp"] = 0
                values["billed-" + pol_name + "-raw-kwh"] = 0
                values["billed-" + pol_name + "-raw-net-gbp"] = 0

            for bill in sess.query(Bill).filter(
                Bill.supply == supply,
                Bill.start_date <= month_finish,
                Bill.finish_date >= month_start,
            ):
                bill_start = bill.start_date
                bill_finish = bill.finish_date
                bill_duration = (bill_finish - bill_start).total_seconds() + 30 * 60
                overlap_duration = (
                    min(bill_finish, month_finish) - max(bill_start, month_start)
                ).total_seconds() + 30 * 60
                overlap_proportion = float(overlap_duration) / float(bill_duration)
                values["billed-import-net-gbp"] += overlap_proportion * float(bill.net)
                values["billed-import-kwh"] += overlap_proportion * float(bill.kwh)

            for era in eras:
                chunk_start = hh_max(era.start_date, month_start)
                chunk_finish = hh_min(era.finish_date, month_finish)

                import_mpan_core = era.imp_mpan_core
                if import_mpan_core is None:
                    continue

                supplier_contract = era.imp_supplier_contract

                if source_code in ["net", "gen-net", "3rd-party"]:
                    supply_source = chellow.computer.SupplySource(
                        sess,
                        chunk_start,
                        chunk_finish,
                        forecast_date,
                        era,
                        True,
                        caches,
                    )

                    values["metered-import-kwh"] += sum(
                        datum["msp-kwh"] for datum in supply_source.hh_data
                    )

                    import_vb_function = supply_source.contract_func(
                        supplier_contract, "virtual_bill"
                    )
                    if import_vb_function is None:
                        values["problem"] += (
                            "Can't find the "
                            "virtual_bill  function in the supplier "
                            "contract. "
                        )
                    else:
                        import_vb_function(supply_source)
                        values["metered-import-net-gbp"] += supply_source.supplier_bill[
                            "net-gbp"
                        ]

                    supply_source.contract_func(era.dc_contract, "virtual_bill")(
                        supply_source
                    )
                    values["metered-import-net-gbp"] += supply_source.dc_bill["net-gbp"]

                    mop_func = supply_source.contract_func(
                        era.mop_contract, "virtual_bill"
                    )
                    if mop_func is None:
                        values["problem"] += (
                            " MOP virtual_bill " "function can't be found."
                        )
                    else:
                        mop_func(supply_source)
                        mop_bill = supply_source.mop_bill
                        values["metered-import-net-gbp"] += mop_bill["net-gbp"]
                        if len(mop_bill["problem"]) > 0:
                            values["problem"] += (
                                " MOP virtual bill problem: " + mop_bill["problem"]
                            )

            values["timestamp"] = int(time.time() * 1000)
            tmp_file.write(
                ",".join('"' + str(values[name]) + '"' for name in field_names) + "\n"
            )


def content(year, month, months, supply_id, user):
    tmp_file = None
    try:
        with Session() as sess:
            if supply_id is None:
                base_name = (
                    f"supplies_monthly_duration_for_all_supplies_for_{months}_to_"
                    f"{year}_{month}.csv"
                )
            else:
                base_name = (
                    f"supplies_monthly_duration_for_{supply_id}_{months}_to_{year}_"
                    f"{month}.csv"
                )
            running_name, finished_name = chellow.dloads.make_names(base_name, user)

            tmp_file = open(running_name, "w")
            _content(sess, tmp_file, year, month, months, supply_id, user)
    except BaseException:
        tmp_file.write(traceback.format_exc())
    finally:
        tmp_file.close()
        os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int("end_year")
    month = req_int("end_month")
    months = req_int("months")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    user = g.user
    threading.Thread(
        target=content, args=(year, month, months, supply_id, user)
    ).start()
    return chellow_redirect("/downloads", 303)



import csv
from decimal import Decimal
from io import StringIO

from chellow.models import (
    BillType,
    Contract,
    GContract,
    GDn,
    GReadType,
    GReadingFrequency,
    GUnit,
    MarketRole,
    Participant,
    Site,
    insert_bill_types,
    insert_g_read_types,
    insert_g_reading_frequencies,
    insert_g_units,
)
from chellow.reports.report_g_supply_virtual_bill import content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_cv_rate_script = {
        "cvs": {
            "EA": {
                1: {"applicable_at": utc_datetime(2020, 10, 3), "cv": 39.2000},
            }
        }
    }
    GContract.insert_industry(sess, "cv", "", {}, vf, None, g_cv_rate_script)
    ug_rate_script = {
        "ug_gbp_per_kwh": {"EA1": Decimal("40.1")},
    }
    GContract.insert_industry(sess, "ug", "", {}, vf, None, ug_rate_script)
    ccl_rate_script = {
        "ccl_gbp_per_kwh": Decimal("0.00525288"),
    }
    GContract.insert_industry(sess, "ccl", "", {}, vf, None, ccl_rate_script)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    charge_script = """
import chellow.gas.ccl
from chellow.gas.engine import g_rates
from chellow.utils import reduce_bill_hhs


def virtual_bill_titles():
    return [
        'units_consumed', 'correction_factor', 'unit_code', 'unit_factor',
        'calorific_value', 'kwh', 'gas_rate', 'gas_gbp', 'ccl_rate',
        'standing_rate', 'standing_gbp', 'net_gbp', 'vat_gbp', 'gross_gbp',
        'problem']


def virtual_bill(ds):
    chellow.gas.ccl.vb(ds)
    for hh in ds.hh_data:
        start_date = hh['start_date']
        bill_hh = ds.bill_hhs[start_date]
        bill_hh['units_consumed'] = hh['units_consumed']
        bill_hh['correction_factor'] = {hh['correction_factor']}
        bill_hh['unit_code'] = {hh['unit_code']}
        bill_hh['unit_factor'] = {hh['unit_factor']}
        bill_hh['calorific_value'] = {hh['calorific_value']}
        kwh = hh['kwh']
        bill_hh['kwh'] = kwh
        gas_rate = float(
            g_rates(ds.sess, ds.caches, db_id, start_date, False)['gas_rate'])
        bill_hh['gas_rate'] = {gas_rate}
        bill_hh['gas_gbp'] = gas_rate * kwh
        bill_hh['ccl_kwh'] = kwh
        ccl_rate = hh['ccl']
        bill_hh['ccl_rate'] = {ccl_rate}
        bill_hh['ccl_kwh'] = kwh
        bill_hh['ccl_gbp'] = kwh * ccl_rate

        bill_hh['aq'] = 4
        bill_hh['soq'] = 5
        if hh['utc_is_month_end']:
            standing_rate = float(
                g_rates(
                    ds.sess, ds.caches, db_id, start_date, False)['standing_rate'])
            bill_hh['standing_rate'] = {standing_rate}
            bill_hh['standing_gbp'] = standing_rate
        if hh['utc_decimal_hour'] == 0:
            pass

        bill_hh['net_gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat_gbp'] = 0
        bill_hh['gross_gbp'] = bill_hh['net_gbp'] + bill_hh['vat_gbp']

    ds.bill = reduce_bill_hhs(ds.bill_hhs)
"""
    g_contract_rate_script = {
        "gas_rate": 0.1,
        "standing_rate": 0.1,
    }
    g_contract = GContract.insert(
        sess, False, "Fusion 2020", charge_script, {}, vf, None, g_contract_rate_script
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    msn = "hgeu8rhg"
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    breakdown = {"units_consumed": 771}
    insert_bill_types(sess)
    bill_type_N = BillType.get_by_code(sess, "N")
    insert_g_read_types(sess)
    g_read_type_A = GReadType.get_by_code(sess, "A")
    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_N,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2015, 9, 1),
        utc_datetime(2015, 9, 30, 22, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )
    g_bill.insert_g_read(
        sess,
        msn,
        g_unit_M3,
        Decimal("1"),
        Decimal("37"),
        Decimal("90"),
        utc_datetime(2015, 9, 1),
        g_read_type_A,
        Decimal("890"),
        utc_datetime(2015, 9, 25),
        g_read_type_A,
    )
    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch(
        "chellow.reports.report_g_supply_virtual_bill.open", return_value=mock_file
    )
    mocker.patch(
        "chellow.reports.report_g_supply_virtual_bill.chellow.dloads.make_names",
        return_value=("a", "b"),
    )
    mocker.patch("chellow.reports.report_g_supply_virtual_bill.os.rename")

    user = mocker.Mock()
    g_supply_id = g_supply.id
    file_name = "a.csv"
    start_date = to_utc(ct_datetime(2020, 9, 1))
    finish_date = to_utc(ct_datetime(2020, 9, 1))

    content(g_supply_id, file_name, start_date, finish_date, user.id)

    mock_file.seek(0)
    sheet = csv.reader(mock_file)
    table = list(sheet)

    expected = [
        [
            "MPRN",
            "Site Code",
            "Site Name",
            "Account",
            "From",
            "To",
            "",
            "units_consumed",
            "correction_factor",
            "unit_code",
            "unit_factor",
            "calorific_value",
            "kwh",
            "gas_rate",
            "gas_gbp",
            "ccl_rate",
            "standing_rate",
            "standing_gbp",
            "net_gbp",
            "vat_gbp",
            "gross_gbp",
            "problem",
        ],
        [
            "87614362",
            "22488",
            "Water Works",
            "d7gthekrg",
            "2020-09-01 00:00",
            "2020-09-01 00:00",
            "",
            "0.6944444444444444",
            "1.0",
            "M3",
            "1.0",
            "39.2",
            "7.561728395061729",
            "0.1",
            "0.7561728395061729",
            "0.00525288",
            "",
            "",
            "0.7958936913580247",
            "0",
            "0.7958936913580247",
            "",
            "aq",
            "4",
            "ccl_gbp",
            "0.039720851851851854",
            "ccl_kwh",
            "7.561728395061729",
            "soq",
            "5",
        ],
    ]

    assert expected == table



import atexit
import collections
import csv
import json
import threading
import traceback
from datetime import datetime as Datetime
from decimal import Decimal
from io import BytesIO

from dateutil.relativedelta import relativedelta


import requests

from sqlalchemy import or_, select
from sqlalchemy.sql.expression import null

from zish import dumps, loads

from chellow.models import Contract, RateScript, Session
from chellow.national_grid import api_get
from chellow.rate_server import download
from chellow.utils import HH, ct_datetime, hh_format, to_ct, to_utc, utc_datetime_now

MAXIMA = [
    {
        "maximum": 15,
        "start": to_utc(ct_datetime(2020, 6, 25)),
        "finish": to_utc(ct_datetime(2020, 8, 31, 23, 30)),
        "reference": "CMP345",
    },
    {
        "maximum": 20,
        "start": to_utc(ct_datetime(2022, 1, 17)),
        "finish": to_utc(ct_datetime(2022, 3, 31, 23, 30)),
        "reference": "CMP381",
    },
    {
        "maximum": 40,
        "start": to_utc(ct_datetime(2022, 10, 6)),
        "finish": to_utc(ct_datetime(2023, 3, 31, 23, 30)),
        "reference": "CMP395",
    },
]


def hh(data_source, run="RF"):
    try:
        bsuos_cache = data_source.caches["bsuos"][run]
    except KeyError:
        try:
            top_cache = data_source.caches["bsuos"]
        except KeyError:
            top_cache = data_source.caches["bsuos"] = {}

        try:
            bsuos_cache = top_cache[run]
        except KeyError:
            bsuos_cache = top_cache[run] = {}

    for h in data_source.hh_data:
        try:
            h["bsuos-rate"] = bsuos_rate = bsuos_cache[h["start-date"]]
        except KeyError:
            h_start = h["start-date"]

            rates = data_source.non_core_rate("bsuos", h_start)

            if h_start >= to_utc(ct_datetime(2023, 4, 1)):
                try:
                    bprice = rates["rate_gbp_per_mwh"]
                except KeyError:
                    bprice = rates["forecast_rate_gbp_per_mwh"]
                bsuos_price = float(bprice)

            else:
                bsuos_rates = rates["rates_gbp_per_mwh"]

                maxi = None
                for max_dict in MAXIMA:
                    if max_dict["start"] <= h_start <= max_dict["finish"]:
                        maxi = max_dict["maximum"]
                        break

                try:
                    bsuos_price_dict = bsuos_rates[key_format(h_start)]
                    bsuos_price = _find_price(run, bsuos_price_dict, maxi)
                except KeyError:
                    ds = bsuos_rates.values()
                    bsuos_price = sum(_find_price(run, d, maxi) for d in ds) / len(ds)

            h["bsuos-rate"] = bsuos_rate = bsuos_cache[h_start] = bsuos_price / 1000

        h["bsuos-kwh"] = h["nbp-kwh"]
        h["bsuos-gbp"] = h["nbp-kwh"] * bsuos_rate


def _find_price(run, prices, maxi):
    try:
        price = prices[run]
    except KeyError:
        try:
            price = prices["RF"]
        except KeyError:
            try:
                price = prices["SF"]
            except KeyError:
                price = prices["II"]

    float_price = float(price)
    return float_price if maxi is None else min(float_price, maxi)


def key_format(dt):
    return dt.strftime("%d %H:%M")


bsuos_importer = None


def _save_cache(sess, cache):
    for yr, yr_cache in cache.items():
        for month, (rs, rates, rts) in tuple(yr_cache.items()):
            rs.script = dumps(rates)
            sess.commit()
            del yr_cache[month]


BASE_URL = "https://data.nationalgrideso.com/backend/dataset/"
PATHS = (
    "d6a4bf54-c63f-4014-a716-49fd3878ca52/resource/"
    "0eda5e28-1dc6-48da-8663-c00e12f2a1e2/download/current_ii_bsuos_data-11.csv",
    "d6a4bf54-c63f-4014-a716-49fd3878ca52/resource/"
    "f0060fd0-1fc9-4288-a0b3-4af9b592b0cf/download/current_sf_bsuos_data.csv",
    "d6a4bf54-c63f-4014-a716-49fd3878ca52/resource/"
    "26b0f410-27d4-448a-9437-45277818b838/download/current_rf_bsuos_data.csv",
)


class BsuosImporter(threading.Thread):
    def __init__(self):
        super(BsuosImporter, self).__init__(name="BSUoS Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=500)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = "proxy.host"
        self.PROXY_PORT_KEY = "proxy.port"

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                self.global_alert = s = None
                with Session() as sess:
                    try:
                        self.log("Starting to check BSUoS rates.")
                        contract = Contract.get_non_core_by_name(sess, "bsuos")
                        props = contract.make_properties()
                        s = requests.Session()
                        s.verify = False
                        if props.get("enabled", False):
                            for path in PATHS:
                                _process_url(
                                    self.log, sess, f"{BASE_URL}{path}", contract, s
                                )
                        else:
                            self.log(
                                "The automatic importer is disabled. To enable it, "
                                "edit the contract properties to set 'enabled' to True."
                            )
                        self.log("Finished checking BSUoS rates.")
                    except BaseException:
                        self.log(f"Outer problem {traceback.format_exc()}")
                        self.global_alert = (
                            f"There's a problem with the "
                            f"<a href='/non_core_contracts/{contract.id}'>BSUoS "
                            f"automatic importer</a>."
                        )
                        sess.rollback()
                    finally:
                        self.lock.release()

            self.going.wait(60 * 60 * 24)
            self.going.clear()


def _process_url(logger, sess, url, contract, s):
    logger(f"Checking to see if there's any new data at {url}")
    sess.rollback()  # Avoid long-running transaction
    res = s.get(url)
    logger(f"Received {res.status_code} {res.reason}")
    cache = {}
    parsed_rows = []

    reader = csv.reader(res.text.splitlines())
    next(reader)  # Skip titles

    for row in reader:
        date_str = row[0].strip()
        date_format = "%Y-%m-%dT%H:%M:%S" if len(date_str) == 19 else "%d-%m-%y"
        date = Datetime.strptime(date_str, date_format)
        period_str = row[1]
        period = int(period_str)
        price_str = row[2]
        price = Decimal(price_str)
        run = row[5]
        parsed_rows.append((date, period, price, run))

    for date, period, price, run in parsed_rows:
        hh_date_ct = to_ct(date)
        hh_date_ct += relativedelta(minutes=30 * (period - 1))
        hh_date = to_utc(hh_date_ct)
        if hh_date >= to_utc(ct_datetime(2023, 4, 1)):
            continue

        try:
            rs, rates, rts = cache[hh_date.year][hh_date.month]
        except KeyError:
            _save_cache(sess, cache)

            try:
                yr_cache = cache[hh_date.year]
            except KeyError:
                yr_cache = cache[hh_date.year] = {}

            rs = (
                sess.query(RateScript)
                .filter(
                    RateScript.contract == contract,
                    RateScript.start_date <= hh_date,
                    or_(
                        RateScript.finish_date == null(),
                        RateScript.finish_date >= hh_date,
                    ),
                )
                .first()
            )
            while rs is None:
                logger(f"There's no rate script at {hh_format(hh_date)}.")
                latest_rs = (
                    sess.query(RateScript)
                    .filter(RateScript.contract == contract)
                    .order_by(RateScript.start_date.desc())
                    .first()
                )
                contract.update_rate_script(
                    sess,
                    latest_rs,
                    latest_rs.start_date,
                    latest_rs.start_date + relativedelta(months=2) - HH,
                    loads(latest_rs.script),
                )
                new_rs_start = latest_rs.start_date + relativedelta(months=1)
                contract.insert_rate_script(sess, new_rs_start, {})
                sess.commit()
                logger(f"Added a rate script starting at {hh_format(new_rs_start)}.")

                rs = (
                    sess.query(RateScript)
                    .filter(
                        RateScript.contract == contract,
                        RateScript.start_date <= hh_date,
                        or_(
                            RateScript.finish_date == null(),
                            RateScript.finish_date >= hh_date,
                        ),
                    )
                    .first()
                )

            rates = loads(rs.script)
            try:
                rts = rates["rates_gbp_per_mwh"]
            except KeyError:
                rts = rates["rates_gbp_per_mwh"] = {}
            yr_cache[hh_date.month] = rs, rates, rts

        key = key_format(hh_date)
        try:
            existing = rts[key]
        except KeyError:
            existing = rts[key] = {}

        if run not in existing:
            existing[run] = price
            logger(f"Added rate at {hh_format(hh_date)} for run {run}.")

    _save_cache(sess, cache)


def get_importer():
    return bsuos_importer


def startup():
    global bsuos_importer
    bsuos_importer = BsuosImporter()
    bsuos_importer.start()


@atexit.register
def shutdown():
    if bsuos_importer is not None:
        bsuos_importer.stop()


def find_rate(file_name, file_like, rate_index):
    rate_script = {"a_file_name": file_name}
    rates = []
    for rate_str in json.load(file_like)["rates"]:
        rates.append(Decimal(rate_str))

    rate_script["rate_gbp_per_mwh"] = rates[rate_index]
    return rate_script


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new BSUoS spreadsheets")

    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year_str, utility, rate_type, file_name = path
            if utility == "electricity" and rate_type == "bsuos":
                year = int(year_str)
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    for year, year_pdfs in sorted(year_entries.items()):
        year_start = to_utc(ct_datetime(year, 4, 1))
        oct_start = to_utc(ct_datetime(year, 10, 1))
        contract = Contract.get_non_core_by_name(sess, "bsuos")
        if year_start < contract.start_rate_script.start_date:
            continue
        rs_1 = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if rs_1 is None:
            rs_1 = contract.insert_rate_script(sess, year_start, {})

        rs_2 = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == oct_start,
            )
        ).scalar_one_or_none()
        if rs_2 is None:
            rs_2 = contract.insert_rate_script(sess, oct_start, {})

        if len(year_pdfs) > 0:
            file_name, url = sorted(year_pdfs.items())[-1]

            rs_1_script = rs_1.make_script()
            if rs_1_script.get("a_file_name") != file_name:
                log(
                    f"Found new file {file_name} for rate script starting "
                    f"{hh_format(year_start)}"
                )
                rs_1.update(find_rate(file_name, BytesIO(download(s, url)), 0))
                log(f"Updated BSUoS rate script for {hh_format(year_start)}")

            rs_2_script = rs_2.make_script()
            if rs_2_script.get("a_file_name") != file_name:
                log(
                    f"Found new file {file_name} for rate script starting "
                    f"{hh_format(oct_start)}"
                )
                rs_2.update(find_rate(file_name, BytesIO(download(s, url)), 1))
                log(f"Updated BSUoS rate script for {hh_format(oct_start)}")

    log("Finished BSUoS spreadsheets")
    sess.commit()


def national_grid_import(sess, log, set_progress, s):
    log("Starting to check for new BSUoS forecast")

    contract = Contract.get_non_core_by_name(sess, "bsuos")

    block = None

    params = {
        "sql": """SELECT * FROM  "578b493e-db5c-41e3-8b52-f91c5f80389c" """
        """ORDER BY "_id" ASC LIMIT 100 """
    }
    res_j = api_get(s, "datastore_search_sql", params=params)
    for record in res_j["result"]["records"]:
        month_start_ct = to_ct(Datetime.strptime(record["Month"], "%b-%y"))
        if month_start_ct.month in (4, 10):
            block = {"start_date": to_utc(month_start_ct), "cost": 0, "vol": 0}

        if block is not None:
            block["cost"] += sum(
                float(record[t])
                for t in (
                    "Balancing Costs (Central) £m",
                    "Estimated Internal BSUoS & ESO Incentive £m",
                    "ALoMCP £m",
                    "CMP381 Deferred Costs £m",
                    "Winter Contingency Cost (Central) £m",
                    "Winter Security of Supply Cost (£m)",
                )
            )
            block["vol"] += float(record["Estimated BSUoS Volume (TWh)"])

            if month_start_ct.month in (3, 9):
                rs = sess.execute(
                    select(RateScript).where(
                        RateScript.contract == contract,
                        RateScript.start_date == block["start_date"],
                    )
                ).scalar_one_or_none()
                if rs is None:
                    rs = contract.insert_rate_script(sess, block["start_date"], {})

                rs_script = rs.make_script()
                rs_script["forecast_rate_gbp_per_mwh"] = block["cost"] / block["vol"]
                rs.update(rs_script)
                sess.commit()

    log("Finished BSUoS forecast")
    sess.commit()



def hh(supply_source):
    try:
        ro_cache = supply_source.caches["ro"]
    except KeyError:
        ro_cache = supply_source.caches["ro"] = {}

    for hh in supply_source.hh_data:
        try:
            hh["ro-rate"] = ro_rate = ro_cache[hh["start-date"]]
        except KeyError:
            h_start = hh["start-date"]
            hh["ro-rate"] = ro_rate = ro_cache[h_start] = float(
                supply_source.non_core_rate("ro", h_start)["ro_gbp_per_msp_kwh"]
            )

        ro_kwh = hh["msp-kwh"]
        hh["ro-msp-kwh"] = ro_kwh
        hh["ro-gbp"] = ro_kwh * ro_rate



import atexit
import csv
import threading
import traceback
from collections import deque
from datetime import datetime as Datetime
from decimal import Decimal

from dateutil.relativedelta import relativedelta


import requests

from sqlalchemy import select


from chellow.models import GContract, GRateScript, Session
from chellow.utils import (
    HH,
    c_months_u,
    hh_after,
    hh_format,
    to_ct,
    to_utc,
    utc_datetime_now,
)


def param_format(dt):
    return to_ct(dt).strftime("%Y-%m-%dT%H:%M:%S")


cv_importer = None


class CvImporter(threading.Thread):
    def __init__(self):
        super().__init__(name="CV Importer")
        self.lock = threading.RLock()
        self.messages = deque(maxlen=5000)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = "proxy.host"
        self.PROXY_PORT_KEY = "proxy.port"
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(f"{hh_format(utc_datetime_now())} - {message}")

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = self.global_alert = g_contract = None
                with Session() as sess:
                    try:
                        g_contract = GContract.get_industry_by_name(sess, "cv")
                        fetch_cvs(sess, self.log, g_contract)
                    except BaseException:
                        self.log(f"Outer problem {traceback.format_exc()}")
                        if g_contract is None:
                            self.global_alert = (
                                "There's a problem with the gas CV automatic importer."
                            )
                        else:
                            self.global_alert = (
                                f"There's a problem with the <a "
                                f"href='/g/industry_contracts/{g_contract.id}'>gas CV "
                                f"automatic importer</a>."
                            )
                        sess.rollback()
                    finally:
                        self.lock.release()
                        self.log("Finished checking CV rates.")

            self.going.wait(60 * 60 * 24)
            self.going.clear()


def fetch_cvs(sess, log_f, g_contract):
    log_f("Starting to check CV rates.")
    props = g_contract.make_properties()
    if not props.get("enabled", False):
        log_f(
            "The automatic importer is disabled. To enable it, edit the contract "
            "properties to set 'enabled' to true."
        )
        return

    now = utc_datetime_now()
    state = g_contract.make_state()
    try:
        last_applicable_at = state["last_applicable_at"]
    except KeyError:
        last_applicable_at = now

    search_start = last_applicable_at - relativedelta(days=1)
    search_finish = now + relativedelta(days=1)

    url = (
        f"https://data.nationalgas.com/api/find-gas-data-download?applicableFor=N&"
        f"dateFrom={param_format(search_start)}&dateTo={param_format(search_finish)}&"
        f"dateType=NORMALDAY&latestFlag=Y&ids=PUBOBJ1660,PUBOB4507,PUBOB4508,PUBOB4510,"
        f"PUBOB4509,PUBOB4511,PUBOB4512,PUBOB4513,PUBOB4514,PUBOB4515,PUBOB4516,"
        f"PUBOB4517,PUBOB4518,PUBOB4519,PUBOB4521,PUBOB4520,PUBOB4522,PUBOBJ1661,"
        f"PUBOBJ1662&type=CSV"
    )
    log_f(
        f"Checking to see if data is was made available from {hh_format(search_start)} "
        f"to {hh_format(search_finish)} at {url}"
    )

    s = requests.Session()
    s.verify = False
    res = s.get(url)
    log_f(f"Received {res.status_code} {res.reason}")

    cf = csv.reader(res.text.splitlines())
    row = next(cf)  # Skip title row
    for row in cf:
        applicable_at_str = row[0]
        applicable_for_str = row[1]
        applicable_for_ct = to_ct(Datetime.strptime(applicable_for_str, "%d/%m/%Y"))
        data_item = row[2]
        value_str = row[3]

        if "LDZ" in data_item:
            ldz = data_item[-3:-1]
            applicable_at = to_utc(
                to_ct(Datetime.strptime(applicable_at_str, "%d/%m/%Y %H:%M:%S"))
            )
            last_applicable_at = max(last_applicable_at, applicable_at)
            cv = Decimal(value_str)
            month_pairs = list(
                c_months_u(
                    start_year=applicable_for_ct.year,
                    start_month=applicable_for_ct.month,
                    months=1,
                )
            )
            month_start, month_finish = month_pairs[0]
            rs = sess.execute(
                select(GRateScript).where(
                    GRateScript.g_contract == g_contract,
                    GRateScript.start_date == month_start,
                )
            ).scalar_one_or_none()
            if rs is None:
                if month_start < g_contract.start_g_rate_script.start_date:
                    rs = g_contract.start_g_rate_script
                    orig_script = g_contract.start_g_rate_script.make_script()
                    g_contract.update_g_rate_script(
                        sess, rs, month_start, rs.finish_date, {"cvs": {}}
                    )
                    g_contract.insert_g_rate_script(
                        sess, month_finish + HH, orig_script
                    )
                elif hh_after(month_start, g_contract.finish_g_rate_script.finish_date):
                    g_contract.update_g_rate_script(
                        sess,
                        g_contract.finish_g_rate_script,
                        g_contract.finish_g_rate_script.start_date,
                        month_finish,
                        g_contract.finish_g_rate_script.make_script(),
                    )
                    rs = g_contract.insert_g_rate_script(sess, month_start, {"cvs": {}})
                else:
                    rs = g_contract.insert_g_rate_script(sess, month_start, {"cvs": {}})

            script = rs.make_script()
            rs_cvs = script["cvs"]
            try:
                ldz_days = rs_cvs[ldz]
            except KeyError:
                ldz_days = rs_cvs[ldz] = {}

            ldz_days[applicable_for_ct.day] = {"applicable_at": applicable_at, "cv": cv}

            g_contract.update_g_rate_script(
                sess, rs, rs.start_date, rs.finish_date, script
            )
            sess.commit()
            log_f(
                f"Added applicable_for: {applicable_for_str}, "
                f"applicable_at: {applicable_at_str}, value: {cv}"
            )

    state["last_applicable_at"] = last_applicable_at
    g_contract.update_state(state)
    sess.commit()


def get_importer():
    return cv_importer


def startup():
    global cv_importer
    cv_importer = CvImporter()
    cv_importer.start()


@atexit.register
def shutdown():
    if cv_importer is not None:
        cv_importer.stop()
